import time
import os
import json
import numpy as np
import subprocess as sp
from typing import Dict
from sklearn.ensemble import RandomForestClassifier
from datasmash.quantizer import Quantizer, vectorize_label
from datasmash.utils import (DatasetLoader, genesess, argmax_prod_matrix_list,
                             pprint_dict)
from datasmash.config import CWD, BIN_PATH
from datasmash._version import __version__

#from d3m_metadata import container, hyperparams, metadata as metadata_module, params, utils
#import d3m.metadata as utils
from d3m.metadata import hyperparams, params, base
from d3m.container import dataset, numpy
from d3m.primitive_interfaces.base import CallResult, DockerContainer
from d3m.primitive_interfaces.supervised_learning import SupervisedLearnerPrimitiveBase


Inputs = dataset.Dataset
Outputs = numpy.ndarray


class Params(params.Params):
    pass


class Hyperparams(hyperparams.Hyperparams):
    """
    a no-op
    """
    depth = hyperparams.UniformInt(
        default = 1000,
        lower = 100,
        upper = 10000,
        description = 'Exponentiated tree depth in inferred Markov Model. ',
        semantic_types=['https://metadata.datadrivendiscovery.org/types/ControlParameter']
    )


class XG2(SupervisedLearnerPrimitiveBase[Inputs, Outputs, Params, Hyperparams]):
    """

    """
    __author__ = "UChicago"
    metadata = base.PrimitiveMetadata({
        "algorithm_types": ['HIDDEN_MARKOV_MODEL', 'RANDOM_WALK',
                            'VARIABLE_ORDER_MARKOV_MODEL'],
        "name": "datasmash.classification.XG1",
        "primitive_family": "TIME_SERIES_CLASSIFICATION",
        "python_path": "d3m.primitives.datasmash.XG1",
        "source": {'name': 'UChicago'},
        "version": __version__,
        "id": "2a1acac4-daa0-4ac0-a543-71956f0681b3",
        'installation': [
            {'type': base.PrimitiveInstallationType.PIP,
             'package': 'datasmash',
             'version': __version__
            }
        ],
        "keywords": [
            'time',
            'series',
            'data smashing',
            'data-smashing',
            'data_smashing',
            'datasmashing',
            'classification',
            'parameter-free',
            'hyperparameter-free'
        ]
    })


    def __init__(self, *,
                 hyperparams: Hyperparams,
                 random_seed: int = 0,
                 docker_containers: Dict[str, DockerContainer] = None,
                 _verbose: int = 0) -> None:

        super().__init__(hyperparams=hyperparams, random_seed=random_seed,
                         docker_containers=docker_containers)

        assert os.path.isfile(os.path.join(BIN_PATH, 'XgenESeSS')), "invalid bin path."
        self._bin_path = BIN_PATH
        self._channel_partition_map = {}
        self._cwd = os.getcwd()
        self._d3m_reader = DatasetLoader()
        self._tmp_dir = ''
        self._channel_dirs = []
        self._selected_channels = set()
        self._detrending = False
        self._inplace = True
        self._channel_probabilities = {}
        self._channel_predictions = {}
        self._fitted = False
        self._depth = self.hyperparams['depth']
        self._channels = None

        #if classifier is not None:
        #    self._classifier = classifier
        #else:
        self._classifier = RandomForestClassifier(n_estimators=500,
                                                  max_depth=None,
                                                  min_samples_split=2,
                                                  random_state=random_seed,
                                                  class_weight='balanced')

    #@property
    #def selected_channels(self):
    #    return self._selected_channels

    #@selected_channels.setter
    #def selected_channels(self, channels):
    #    if not isinstance(channels, list):
    #        channels_ = [channels]
    #    else:
    #        channels_ = channels
    #    channels_ = ['channel_' + str(c) for c in channels_]
    #    self._selected_channels = set(channels_)

    def set_training_data(self, *, inputs: Inputs, outputs: Outputs) -> None:
        """
        outputs argument should be specified as None
        """
        self._d3m_reader.load_dataset(data=inputs,
                                      train_or_test='train')
        self._tmp_dir, self._channel_dirs, self.channel_problems, self.y = (
            self._d3m_reader.write_libs(problem_type='supervised'))
        self._fitted = False

    def _fit_one_channel(self, directory):
        """

        """
        y = vectorize_label(self.channel_problems['channel_0']['raw_libs'], self.y)
        qtz = Quantizer(problem_type='supervised', multi_partition=True,
                        featurization=genesess,
                        featurization_params={'multi_line': True, 'depth': self._depth})
        X = qtz.fit_transform(directory)
        return X, y, qtz

    def fit(self, *, timeout: float = None, iterations: int = None) -> CallResult[None]:
        """

        """
        channels = self._channels
        if channels is not None:
            if channels == -1:
                pass
            elif not isinstance(channels, list):
                channels = [channels]
            selected_channels = ['channel_' + str(c) for c in channels]
            self._selected_channels = set(selected_channels)
        elif bool(self._selected_channels):
            selected_channels = self._selected_channels
        else:
            selected_channels = self._channel_dirs

        for channel_dir in selected_channels:
            X, self.y, qtz = self._fit_one_channel(channel_dir) # TODO: self.y
            channel_name = os.path.basename(channel_dir)
            self._channel_partition_map[channel_name] = qtz

        #if verbose:
        #    print('Quantizing in place:', self.inplace)
        #    print('Chosen partition:')
        #    pprint_dict(self._channel_partition_map)

        #self._classifier.fit(X, self.y)

        self._fitted = True
        return CallResult(None)

    def produce(self, *, inputs: Inputs, timeout: float = None,
                iterations: int = None) -> CallResult[Outputs]:
        """

        """
        #if partition is not None:
            #    current_partition = partition
            #elif self._channel_partition_map is not None:
                #current_partition = self._channel_partition_map
                #self._d3m_reader.load_dataset(data=inputs,
                #                              train_or_test='test')
        current_partition = self._channel_partition_map
        self._d3m_reader.load_dataset(data=inputs,
                                      train_or_test='test')

        channels = self._channels
        if channels is not None:
            if not isinstance(channels, list):
                channels = [channels]
                if bool(self._selected_channels):
                    for channel in channels:
                        if channel not in self._selected_channels:
                            raise ValueError("The partition was not found for this "
                                             "channel. Re-run 'fit()' with this "
                                             "channel included before running "
                                             "'produce()' with this channel.")

        channel_problems = self._d3m_reader.write_test()

        for channel, problem in channel_problems.items():
            if channels is not None:
                channels_ = ['channel_' + str(c) for c in channels]
                if channel not in channels_:
                    #if verbose:
                        print('Excluding', channel, '...')
                        continue
                elif bool(self._selected_channels) and (channel not in
                                                        self._selected_channels):
                    #if verbose:
                        print('Excluding', channel, '...')
                        continue

            #if verbose:
            #start = time.time()
            test_file = problem['test']


            qtz = current_partition[channel]
            #print(json.dumps(qtz.parameters, indent=4))
            #print(json.dumps(qtz.data, indent=4))

            X = qtz.transform(test_file)
            print(X.shape)
            print(qtz.training_X.shape)
            self._classifier.fit(qtz.training_X, self.y)
            predictions = self._classifier.predict(X)
            return CallResult(predictions)
            #quantize_inplace(test_file, current_partition[channel],
            #                 pruning=self._prune_range,
            #                 detrending=self._detrending,
            #                 normalization=self._normalization)

            #test_features = line_by_line(test_file, function=xgenesess,
            #                             max_delay=self._max_delay)

            #X = np.loadtxt(test_features)
            #print(X.shape)
            #probabilities = self._classifier.predict_proba(X)
            #predictions = self._classifier.predict(X)
            #self._channel_probabilities[channel] = probabilities
            #self._channel_predictions[channel] = predictions
            #print(predictions)

            ##if verbose:
                #print('CHANNEL ' + channel.split('_')[-1] + ' DONE')
                #print(predictions)
                #end = time.time()
                #print('TIME:', end - start, '\n')
        #prob_list = list(self._channel_probabilities.values())
        #overall_predictions = argmax_prod_matrix_list(prob_list,
        #                                              index_class_map=self._d3m_reader.index_class_map)
        #return CallResult(overall_predictions)

    #@property
    #def channel_predictions(self):
    #    return self._channel_predictions


    def get_params(self) -> Params:
        """
        A noop
        """
        return None

    def set_params(self, *, params: Params) -> None:
        """
        A noop
        """
        return None

